/* Copyright (C) 2002 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: tbdcnv.c,v $
   Revision 1.2  2003/05/03 19:27:32  wb6ymh
   Added permute_bytes() courtesy of Paul Williamson kb5mu to correct endian
   issues on big endian machines (tested by Paul on Mac G4 running OS X).

   Revision 1.1.1.1  2002/11/17 18:46:21  wb6ymh
   Initial import.


*/

#include "common.h"

#ifndef _WIN32
   // FreeBSD, Linux, etc..
   #ifdef HAVE_UNISTD_H
      #include <unistd.h>
   #endif
   
   #ifdef STDC_HEADERS
      #include <stdlib.h>
   #endif
   
   #ifdef TIME_WITH_SYS_TIME
      #include <sys/time.h>
      #include <time.h>
   #else
      #ifdef HAVE_SYS_TIME_H
         #include <sys/time.h>
      #else
         #include <time.h>
      #endif
   #endif
   #include <errno.h>
   
   #ifdef HAVE_STRINGS_H
      #include <string.h>
   #endif
	#include <assert.h>
#else
   // Windoze
   #include <stdio.h>
   #include <stdlib.h>
   #include <string.h>
   #include <time.h>
   #include <sys\timeb.h>
   #include <io.h>
   #include <winsock2.h>
#endif

#include "tbd.h"
#include "ilink.h"
#include "rtp.h"
#include "gsm.h"

#define CONF_BUF_SIZE   1024
#define WAVE_EXT        ".wav"
#define RIFF_STRING     "RIFF"
#define WAVE_STRING     "WAVE"
#define FMT_STRING      "fmt "
#define WAVEFMT_STRING  WAVE_STRING FMT_STRING
#define DATA_STRING     "data"

#define WriteRiffInt(x,y)           \
   fputc(((y) & 0x000000ff),x);     \
   fputc(((y) & 0x0000ff00) >> 8,x);   \
   fputc(((y) & 0x00ff0000) >> 16,x);  \
   fputc(((y) & 0xff000000) >> 24,x)
   
#define WriteRiffShort(x,y)            \
   fputc(((y) & 0x000000ff),x);     \
   fputc(((y) & 0x0000ff00) >> 8,x)

// Use a non-case sensitive compare function if one is available
#ifdef HAVE_STRICMP
   #define STRCMP stricmp
#else
#ifdef HAVE_STRCASECMP
   #define STRCMP strcasecmp
#else
   #define STRCMP strcmp
#endif
#endif

#ifndef	FALSE
#define	FALSE		0
#endif

#ifndef	TRUE
#define	TRUE	!FALSE
#endif

enum {
   ERR_DUP_WAVE_OUTPUTS = 1,
   ERR_DUP_TXT_OUTPUTS,
   ERR_CANT_OPEN_INPUT,
   ERR_CANT_OPEN_OUTPUT,
   ERR_NO_WAV2WAV,
   ERR_NO_OUTPUT,
   ERR_NO_INPUT,
   ERR_DISK_READ_ERR,
   ERR_DISK_WRITE_ERR,
   ERR_INPUT_FORMAT,
   ERR_USAGE_REQUEST,
   ERR_INVALID_WAVEFILE
} EXIT_ERROR_CODES;

char *OutputFilename = NULL;
char *OutputWaveFilename = NULL;
char *InputFilename = NULL;
int bInputIsWav = FALSE;

int TimeStamps = 0;
int DebugLevel = 0;
int Quiet = 0;
int Verbose = 0;
int FileSize = 0;

char *TimeT2String(time_t time);
int IsWav(char *Filename);
int CheckWavHeader(FILE *fp);
int CheckString(FILE *fp,char *TestString);
int ReadRiffInt(FILE *fp,int *Value);

#ifndef _WIN32
void permute_bytes(void *buf,size_t len);
#endif

#ifndef HAVE_GETOPT
// globals needed by getopt
int optind = 1;
int optopt;
char *optarg;
int getopt(int argc,char **argv,char *opts);
#endif

void Usage()
{
   printf("Usage: tbdcnv [-t][-q][-v] -i <input file> -o <output file> "
          "[-o <output file>]\n");
   printf("\t-i - set Input filename.\n");
   printf("\t-o - set Output filename.\n");
   printf("\t-q - Quiet.\n");
   printf("\t-t - add Timestamps to text file output.\n");
   printf("\t-v - verbose status output.\n\n");
   printf("tbdcnv converts files to and from the internal format used\n");
   printf("by thebridge.\n\n");
   printf("If the output file has an extension of \".wav\" then the audio\n");
   printf("from thebridge input file is extracted.\n\n");
   printf("If the input file has an extension of \".wav\" then the\n");
   printf("audio is converted into a output file in thebridge format.\n\n");
   printf("If the neither then input file nor the output file has an extension\n");
   printf("of \".wav\" then the input file is assumed to be in the thebridge\n");
   printf("format and the text messages are extracted to the output file.\n\n");
   printf("REQUIRED Wave file format:\n");
   printf("\tPCM (uncompressed)\n");
   printf("\t16 bit samples\n");
   printf("\tmono\n");
   printf("\t8000 Hz sampling rate\n");
}

int main(int argc, char* argv[])
{
   FILE *fin = NULL;
   FILE *OutFp = NULL;
   FILE *wavFp = NULL;
   FilePacketHdr PktHdr;
   char TempBuf[CONF_BUF_SIZE];
   gsm_frame gsmout;
   char *cp;
   gsm gsmh = NULL;
   gsm_signal dec[160];
   gsm_byte *pGSM;
   int Option;
   int i;
   long DataSizeOffset = 0;
   long FileSizeOffset = 0;
   long TotalFileSize;
   time_t FirstTimeStamp = 0;
   time_t LastTimeStamp = 0;
   int Records = 0;
   int Ret = 0;
   rtp_hdr_t RTP;
   rtp_hdr_t *pRTP = (rtp_hdr_t *) TempBuf;
   u_int16 Sequence = 1;
   int DataLen;
   long x;
   int bFirstPacket = FALSE;
   unsigned short Seq;
   unsigned short NextSeq = 0;

   while((Option = getopt(argc, argv, "qo:i:tv")) != -1 && !Ret) {
      switch(Option) {
         case 'q':   // quiet 
            Quiet = TRUE;
            break;

         case 'i':   // input filename 
            if(InputFilename != NULL) {
               free(InputFilename);
            }
            InputFilename = strdup(optarg);
            if(IsWav(InputFilename)) {
               bInputIsWav = TRUE;
            }
            break;
         
         case 'o':   // output filename
            if(IsWav(optarg)) {
               if(OutputWaveFilename != NULL) {
                  printf(PACKAGE ": Error - Wavefile output already specified.\n");
                  Usage();
                  Ret = ERR_DUP_WAVE_OUTPUTS;
               }
               OutputWaveFilename = strdup(optarg);
            }
            else {
               if(OutputFilename != NULL) {
                  printf(PACKAGE ": Error - text output already specified.\n");
                  Usage();
                  Ret = ERR_DUP_TXT_OUTPUTS;
               }
               OutputFilename = strdup(optarg);
            }
            break;
         
         case 't':
            TimeStamps = TRUE;
            break;

			case 'v':
				Verbose = TRUE;
				break;

         case '?':
            Usage();
            Ret = ERR_USAGE_REQUEST;
      }
   }

   if(!Quiet && !Ret) {
      printf(PACKAGE " Version " VERSION "\n");
   }
   
   if(InputFilename == NULL) {
      printf(PACKAGE ": Error - input file not specified.\n");
      Usage();
      Ret = ERR_NO_INPUT;
   }
   else if(OutputFilename == NULL && OutputWaveFilename == NULL) {
      printf(PACKAGE ": Error - output file not specified.\n");
      Usage();
      Ret = ERR_NO_OUTPUT;
   }
   else if(bInputIsWav && OutputWaveFilename != NULL) {
      printf(PACKAGE ": Error - can't convert from .wav to .wav.\n");
      Usage();
      Ret = ERR_NO_WAV2WAV;
   }

   if(!Ret && (fin = fopen(InputFilename,MODE_RD_BIN)) == NULL) {
      printf(PACKAGE ": Error - unable to open input file \"%s\".\n",
             InputFilename);
      Ret = ERR_CANT_OPEN_INPUT;
   }

   if(!Ret && OutputFilename != NULL) {
      if(bInputIsWav) {
      // Input file is a Wave file then output is thebridge format
         cp = MODE_WRITE_BIN;
      }
      else {
      // Input file is thebridge format, output file is text
         cp = MODE_WRITE_TXT;
      }
      if((OutFp= fopen(OutputFilename,cp)) == NULL) {
         printf(PACKAGE ": Error - unable to open output file \"%s\".\n",
                OutputFilename);
         Ret = ERR_CANT_OPEN_OUTPUT;
      }
   }
   
   if(!Ret && bInputIsWav) {
      gsmh = gsm_create();
   // Check wave header
      Ret = CheckWavHeader(fin);

      RTP.version = ILINK_RTP_VERSION;
      RTP.p = 0;
      RTP.x = 0;
      RTP.cc = 0;
      RTP.m = 0;
      RTP.pt = 3;   // GSM
      RTP.ts = 0;
      RTP.ssrc = 0;

      while(!Ret && !feof(fin)) {
         if((Ret = CheckString(fin,DATA_STRING)) != 0) {
            if(Ret == ERR_INVALID_WAVEFILE) {
            // Not really invalid, just not a data chunk, ignore it
               if((Ret = ReadRiffInt(fin,&DataLen)) != 0) {
                  break;
               }
               fseek(fin,DataLen,SEEK_CUR);
               continue;
            }
            else {
            // Disk read error
               break;
            }
         }
			if(feof(fin)) {
				break;
			}
         
			if((Ret = ReadRiffInt(fin,&DataLen)) != 0) {
            break;
         }

         while(DataLen > 0) {
         // write header
            PktHdr.Flags = htonl(FILE_FLAG_RATE_LIMIT);
            PktHdr.Len = htonl(ILINK_RTP_FRAME_LEN);
            if(fwrite(&PktHdr,sizeof(PktHdr),1,OutFp) != 1) {
               Ret = ERR_DISK_WRITE_ERR;
               break;
            }

         // write EchoLink (modified RTP) header

            RTP.seq = htons(Sequence++);
            if(fwrite(&RTP,RTP_HDR_SIZE,1,OutFp) != 1) {
               Ret = ERR_DISK_WRITE_ERR;
               break;
            }
            
            for(i = 0; i < 4; i++) {
               if(DataLen < sizeof(dec)) {
               // Not enough data left for another packet, zero fill data
                  memset(dec,0,sizeof(dec));
                  if(DataLen > 0) {
                  // Skip the rest of the data in this chunk
                     fseek(fin,DataLen,SEEK_CUR);
                     DataLen = 0;
                  }
               }
               else {
                  if(fread(dec,sizeof(dec),1,fin) != 1) {
                     perror(PACKAGE ": Error - reading wave data, ");
                     Ret = ERR_DISK_READ_ERR;
                     break;
                  }
                  DataLen -= sizeof(dec);
               }

#ifndef _WIN32
					permute_bytes(dec, sizeof(dec));
#endif
               gsm_encode(gsmh,dec,gsmout);
               if(fwrite(gsmout,sizeof(gsmout),1,OutFp) != 1) {
                  Ret = ERR_DISK_WRITE_ERR;
                  break;
               }
            }
         }
      }
      if(!Ret) {
      // We're done
         PktHdr.Flags = 0;
         PktHdr.Len = 0;
         if(fwrite(&PktHdr,sizeof(PktHdr),1,OutFp) != 1) {
            Ret = ERR_DISK_WRITE_ERR;
         }
      }
   }
   else if(!Ret) {
      if(OutputWaveFilename != NULL) {
         if((wavFp = fopen(OutputWaveFilename,MODE_WRITE_BIN)) == NULL) {
            printf(PACKAGE ": Error - unable to open output text file \"%s\".\n",
                   OutputWaveFilename);
            Ret = ERR_CANT_OPEN_OUTPUT;
         }
         else {
            gsmh = gsm_create();
         // Write wave header
            fprintf(wavFp,RIFF_STRING);
            FileSizeOffset = ftell(wavFp);
            WriteRiffInt(wavFp,0);     /* File size, we'll fill it in later */
            fprintf(wavFp,WAVEFMT_STRING);
            WriteRiffInt(wavFp,16);    /* Chunk size */
            WriteRiffShort(wavFp,1);   /* Format tag - uncompressed */
            WriteRiffShort(wavFp,1);   /* Channels = 1 */
            WriteRiffInt(wavFp,8000);  /* Sample frequency = 8000 Hz */
            WriteRiffInt(wavFp,16000); /* Average bytes per second */
            WriteRiffShort(wavFp,2);   /* Block alignment */
            WriteRiffShort(wavFp,16);  /* Bits per sample */
            fprintf(wavFp,DATA_STRING);
            DataSizeOffset = ftell(wavFp);
            WriteRiffInt(wavFp,0);     /* Data size, we'll fill it in later */
         }
      }

      for( ; ; ) {
         if(fread(&PktHdr,sizeof(PktHdr),1,fin) != 1) {
            perror(PACKAGE ": Error - reading header, ");
            Ret = ERR_DISK_READ_ERR;
            break;
         }
         Records++;
         if(!Quiet && Records % 1000 == 0) {
            printf(".");
            fflush(stdout);
         }

         x = ftell(fin);
         PktHdr.Flags = ntohl(PktHdr.Flags);
         PktHdr.Len = ntohl(PktHdr.Len);

         if(PktHdr.Flags & FILE_FLAG_TIME_STAMP) {
            LastTimeStamp = (time_t) PktHdr.Len;
            if(FirstTimeStamp == 0) {
               FirstTimeStamp = LastTimeStamp;
               if(FirstTimeStamp != 0) {
                  printf("Input file was recorded on %s.\n",
                         TimeT2String(LastTimeStamp));
               }
            }
            printf("%s:\n",TimeT2String(LastTimeStamp));
            bFirstPacket = TRUE;
            continue;
         }

         if(PktHdr.Len == 0) {
            break;
         }

         if(PktHdr.Len > CONF_BUF_SIZE || PktHdr.Len < 0) {
            printf(PACKAGE ": Invalid packet len %d\n",PktHdr.Len);
            Ret = ERR_INPUT_FORMAT;
            break;
         }

         if(fread(TempBuf,PktHdr.Len,1,fin) != 1) {
            perror(PACKAGE ": reading file, ");
            Ret = ERR_DISK_READ_ERR;
            break;
         }

         if(PktHdr.Flags & FILE_FLAG_DATA_PKT) {
         // Data packet
            if((cp = strchr(TempBuf,'\r')) != 0) {
               *cp = 0;
            }
            if(OutFp != NULL) {
               if(TimeStamps) {
                  fprintf(OutFp,"%s:\n",TimeT2String(LastTimeStamp));
               }
               fprintf(OutFp,"%s\n",&TempBuf[6]);
            }
         }
         else {
         // Audio packet
            Seq = ntohs(pRTP->seq);
            if(bFirstPacket) {
               bFirstPacket = FALSE;
               NextSeq = Seq;
            }
            else {
               if(Seq != NextSeq && Verbose) {
                  printf("Missed %d packets, Expected %d, got %d\n",
                         Seq-NextSeq,NextSeq,Seq);
                  NextSeq = Seq;
               }
            }
            NextSeq++;

            if(wavFp != NULL) {
               for(i = RTP_HDR_SIZE; i < PktHdr.Len; i += sizeof(gsm_frame)) {
                  pGSM = (gsm_byte *) &TempBuf[i];
						gsm_decode(gsmh,pGSM, dec);
#ifndef _WIN32
                  permute_bytes(dec, sizeof(dec));
#endif
                  if(fwrite(dec,sizeof(dec),1,wavFp) != 1) {
                     Ret = ERR_DISK_WRITE_ERR;
                     perror("audio file write failed ");
                     break;
                  }
               }
            }
         }
      }

      if(wavFp != NULL) {
      // update header with final values
         TotalFileSize = ftell(wavFp) - 8;
         fseek(wavFp,FileSizeOffset,SEEK_SET);
         WriteRiffInt(wavFp,TotalFileSize);  /* File size */
         fseek(wavFp,DataSizeOffset,SEEK_SET);
         WriteRiffInt(wavFp,TotalFileSize-36);  /* Data size */
      }
   }

   if(fin != NULL) {
      fclose(fin);
   }

   if(wavFp != NULL) {
      fclose(wavFp);
   }

   if(OutFp != NULL) {
      fclose(OutFp);
   }

   if(InputFilename != NULL) {
      free(InputFilename);
   }

   if(OutputWaveFilename != NULL) {
      free(OutputWaveFilename);
   }

   if(OutputFilename!= NULL) {
      free(OutputFilename);
   }
   
   if(!Ret) {
      printf("\n");
      printf("Complete.\n");
   }

   if(Ret == ERR_USAGE_REQUEST) {
      Ret = 0;
   }
   else if(Ret == ERR_DISK_WRITE_ERR) {
      perror(PACKAGE "Error writing disk, ");
   }
   return Ret;
}

char *TimeT2String(time_t time)
{
   struct tm *pTm;
   static char TimeString[] = "mm/dd hh:mm:ss";

   if(time != 0) {
      pTm = localtime(&time);
      if(pTm != NULL) {
         snprintf(TimeString,sizeof(TimeString),"%d/%02d %d:%02d:%02d",
                  pTm->tm_mon + 1,pTm->tm_mday,pTm->tm_hour,pTm->tm_min,
                  pTm->tm_sec);
      }
      else {
         return "";
      }
   }
   else {
      return "";
   }
   return TimeString;
}

int CheckString(FILE *fp,char *TestString)
{
   char TempBuf[20];
   int StrLen = strlen(TestString);

   if(fread(TempBuf,StrLen,1,fp) != 1) {
		if(!feof(fp)) {
			return ERR_DISK_READ_ERR;
		}
		return 0;
   }

   if(strncmp(TempBuf,TestString,StrLen) != 0) {
      return ERR_INVALID_WAVEFILE;
   }
   
   return 0;
}


int ReadRiffInt(FILE *fp,int *Value)
{
   unsigned char x;
   int y;
   int i;

   *Value = 0;
   for(i = 0; i < 4; i++) {
      if(fread(&x,sizeof(x),1,fp) != 1) {
         return ERR_DISK_READ_ERR;
      }
      y = x;
      y <<= 8 * i;
      *Value += y;
   }

   return 0;
}

int ReadRiffShort(FILE *fp,short *Value)
{
   unsigned char x;
   int y;
   int i;

   *Value = 0;
   for(i = 0; i < 2; i++) {
      if(fread(&x,sizeof(x),1,fp) != 1) {
         return ERR_DISK_READ_ERR;
      }
      y = x;
      y <<= 8 * i;
      *Value += y;
   }

   return 0;
}

struct {
   char *Type;
   int   Value;
   char *String;
   char *ErrorMsg;
} HeaderCheck[] = {
   {"c",0,RIFF_STRING,"not a RIFF file"},
   {"if",0,NULL,NULL},     // filesize
   {"c",0,WAVE_STRING,"not a WAVE file"},
   {"c",0,FMT_STRING,"fmt chunk not found"},
   {"ic",0,NULL},    // chunk size
   
   // Format tag - uncompressed
   {"s",1,NULL,"compressed file, must be uncompressed"}, 
   
   // Channels - 1
   {"s",1,NULL,"stereo file, must be mono"},    

   // Sample frequency = 8000 Hz
   {"i",8000,NULL,"incorrect samping rate, must be 8000 Hz"},  
   
   // Average bytes per second = 16000
   {"i",16000,NULL,"bytes per second is incorrect, must be 16000"},  
   
   {"sx",0,NULL},    // Block alignment

   // Bits per sample = 16
   {"s",16,NULL,"incorrect sample size, must be 16 bit"},

   {"."}             // That's it
};

int CheckWavHeader(FILE *fp)
{
   int i;
   int x;
   short y;
   int Ret = 0;
   int ChunkLen = 0;

   for(i = 0; HeaderCheck[i].Type[0] != '.'; i++) { 
      switch(HeaderCheck[i].Type[0]) {
         case 'i':   // int
            ChunkLen -= 4;
            if((Ret = ReadRiffInt(fp,&x)) == 0) {
               if(HeaderCheck[i].Type[1] == 'c') {
               // Chunk Len 
                  ChunkLen = x;
               }
               else if(HeaderCheck[i].Type[1] == 'f') {
               // File size 
                  FileSize = x;
               }
               else if(HeaderCheck[i].Type[1] != 'x') {
                  if(HeaderCheck[i].Value != x) {
                     Ret = ERR_INVALID_WAVEFILE;
                  }
               }
            }
            break;

         case 's':   // short
            ChunkLen -= 2;
            if((Ret = ReadRiffShort(fp,&y)) == 0 && 
               HeaderCheck[i].Type[1] != 'x' &&
               HeaderCheck[i].Value != y)
            {
               Ret = ERR_INVALID_WAVEFILE;
            }
            break;

         case 'c':   // character string
            ChunkLen -= strlen(HeaderCheck[i].String);
            Ret = CheckString(fp,HeaderCheck[i].String);
            break;
      }
      if(Ret) {
         break;
      }
   }

   if(!Ret && ChunkLen > 0) {
   // Skip rest of fmt chunk we don't understand
      fseek(fp,ChunkLen,SEEK_CUR);
   }
   
   if(Ret == ERR_INVALID_WAVEFILE) {
      if(HeaderCheck[i].ErrorMsg != NULL) {
         printf(PACKAGE ": Error - %s\n",HeaderCheck[i].ErrorMsg); 
      }
      else {
         printf(PACKAGE ": Unknown error\n"); 
      }
   }
   else if(Ret == ERR_DISK_READ_ERR) {
      perror(PACKAGE ": Error - reading wave header, ");
   }

   return Ret;
}

int IsWav(char *Filename)
{
   int Ret = FALSE;
   char *cp = strrchr(Filename,'.');

   if(cp != NULL) {
      if(STRCMP(cp,WAVE_EXT) == 0) {
         Ret = TRUE;
      }
   }

   return Ret;
}

#ifndef _WIN32
void permute_bytes(void *buf,size_t len)
{
	gsm_signal *wptr = buf;
	gsm_signal sig;
	char       *bptr = buf;  
	size_t		i;

	assert((len & 0x01) == 0);
	assert(sizeof(gsm_signal) == 2);

	for(i=0; i < len; i += sizeof(gsm_signal)) {
		sig = *wptr++;
		*bptr++ = sig & 0x00FF;
		*bptr++ = (sig >> 8) & 0x00FF;
	}
}
#endif

#ifndef HAVE_GETOPT

/*
 * getopt a wonderful little function that handles the command line.
 * available courtesy of AT&T.
 */
int getopt(int argc,char **argv,char *opts)
{
   static int sp = 1;
   register int c;
   register char *cp;

   if(sp == 1)
      if(optind >= argc ||
         argv[optind][0] != '-' || argv[optind][1] == '\0')
         return(EOF);
      else if(strcmp(argv[optind], "--") == 0) {
         optind++;
         return(EOF);
      }
   optopt = c = argv[optind][sp];
   if(c == ':' || (cp=strchr(opts, c)) == NULL) {
      printf(PACKAGE ": illegal option -- ", c);
      if(argv[optind][++sp] == '\0') {
         optind++;
         sp = 1;
      }
      return('?');
   }
   if(*++cp == ':') {
      if(argv[optind][sp+1] != '\0')
         optarg = &argv[optind++][sp+1];
      else if(++optind >= argc) {
         printf(PACKAGE ": option requires an argument -- ", c);
         sp = 1;
         return('?');
      }
      else
         optarg = argv[optind++];
      sp = 1;
   }
   else {
      if(argv[optind][++sp] == '\0') {
         sp = 1;
         optind++;
      }
      optarg = NULL;
   }
   return(c);
}
#endif


